/**
 * 文件依赖的合并
 */
var path = require('path');
var fs = require('fs');


module.exports = function(grunt){
	
	var jsFileReg = /^[^\.].*\.js$/;
	var defineNameReg = /define\(\s*["'](.*?)['"]/;
	var requireReg = /\brequire\(\s*['"]([^'"\s]+)['"]\s*\)/g;
	var containsReg = /[^\/]\/\/\/contains\(['"](.*?)['"]\)/g;
	
    grunt.registerMultiTask('weibo-compiler', 'compiler .js files', function(){
        var options = this.options();
        var excModule = options.excModule;
        var makeDeps = options.makeDeps;
		
        this.files.forEach(function(file) {
        
            var moduleMap = {};
            var filepathMap = {};
            
            function forIt(filepath, fn) {
                if (fs.statSync(filepath).isDirectory()) {
                    fs.readdirSync(filepath).forEach(function(item){
                        forIt(path.join(filepath, item), fn);
                    });
                } else {
					if (jsFileReg.test(filepath)) {
						fn(filepath.replace(/\\/g, '/'));
					}
                }
            }
            //读出目录下所有文件 并组织成模块数据对象
            forIt(file.cwd, function(filepath) {
                var src = grunt.file.read(filepath);
                var moduleId = src.match(defineNameReg)[1];
				
                filepathMap[filepath] = moduleId;
                moduleMap[moduleId] = {
                    filepath: filepath,
                    src: src
                };
            });
            
            
            //取得所有文件的直接依赖
            for (var moduleId in moduleMap) {
                requireModuleDeps(moduleId, moduleMap, excModule, makeDeps, file.cwd);
            }
            
            
            file.src.map(function(filepath) {
                var moduleId = filepathMap[path.join(file.cwd, filepath).replace(/\\/g, '/')];
                if (moduleId) {
                    var requiredMap = {};
                    var result = concatModule(moduleId, moduleMap, excModule, file.cwd);
                    grunt.file.write(moduleMap[moduleId].filepath, result);
                    console.log('concat module ' + moduleId);
                }
            });
            
            
        });
    });
    
    /**
     * 合并入口方法
     * @param {Object} moduleId
     * @param {Object} moduleMap
     * @param {Object} excModule
     * @param {Object} cwd
     */
    function concatModule(moduleId, moduleMap, excModule){
        //取出模块内的自定义依赖
        var deps = requireRegModuleDeps(moduleId, moduleMap, excModule);
        var ret = [];
        //合并
        deps.forEach(function(mid) {
            ret.push(moduleMap[mid].src);
        });
        return ret.join('\n');
    }
    
    function requireModuleDeps(moduleId, moduleMap, excModule, makeDeps, cwd) {
    
        var moduleObj = moduleMap[moduleId];
        if (!moduleObj) {
            throw Error('moduleId(' + moduleId + ') requireModuleDeps error: module is not defined!');
        }
        if (moduleObj.deps) {
            return moduleObj.deps;
        }
        var directdeps = ['require','exports','module'];
        moduleObj.deps = {};
        var fileFullPath = moduleObj.filepath;
        var fileFolderPath = path.dirname(fileFullPath);
        
        removeComment(moduleObj.src).replace(requireReg, function(_, moduleId) {
			
			var oldmoduleId = moduleId;
			
            if (/^\./.test(moduleId)) {
                moduleId = path.join(fileFolderPath, moduleId).replace(/\\/g, '/').replace(new RegExp('^' + cwd), '').replace(/^[\\\/]/, '');
            }
            if (excModule.indexOf(moduleId) != -1) {
                return;
            }
			directdeps.push(oldmoduleId);
            moduleObj.deps[moduleId] = 1;
            var _deps = requireModuleDeps(moduleId, moduleMap, excModule, makeDeps, cwd);
            for (var mid in _deps) {
                moduleObj.deps[mid] = 1;
            }
        });
		
		if (makeDeps) {
			moduleObj.src = moduleObj.src.replace(/(define\(["'].*?['"],)/, '$1 ' + JSON.stringify(directdeps) + ',');
		}
		
        return moduleObj.deps;
    }
    
    
    function requireRegModuleDeps(moduleId, moduleMap, excModule) {
        var moduleObj = moduleMap[moduleId];
        if (!moduleObj) {
            throw 'moduleId(' + moduleId + ') requireRegModuleDeps error: module is not defined!';
        }
        var directdeps = moduleObj.directdeps;
		var deps = moduleObj.deps;
		
        var moduleMiddleSrc = [moduleObj.src];
        var depsMap = {};
        for (var mid in deps) {
            moduleMiddleSrc.push(moduleMap[mid].src);
            depsMap[mid] = 1;
        }
        moduleMiddleSrc = moduleMiddleSrc.join('\n');
        
        var regList = [];
        
        removeComment(moduleMiddleSrc, {
            block: 1
        }).replace(containsReg, function(_, moduleReg) {
            regList.push('(' + moduleReg.replace(/^\^/, '') + ')');
        });
        
        var reg = regList.length && new RegExp('^' + regList.join('|'));
        if (reg) {
            
            for (var _moduleId in moduleMap) {
                if (_moduleId != moduleId && excModule.indexOf(_moduleId) == -1 && reg.test(_moduleId)) {
                    depsMap[_moduleId] = 1;
                    for (var mid in moduleMap[_moduleId].deps) {
                        depsMap[mid] = 1;
                    }
                }
            }
        }
        var allDeps = [moduleId];
        
        for (var mid in depsMap) {
            allDeps.push(mid);
        }
        
        return allDeps;
    }
    
    function removeComment(src, options){
        options = options ||
        {
            line: true,
            block: true
        };
        
        if (options.line) {
            src = src.replace(/(?:^|\n|\r)\s*\/\/.*(?:\r|\n|$)/g, '\n');
        }
        if (options.block) {
            src = src.replace(/(?:^|\n|\r)\s*\/\*[\s\S]*?\*\/\s*(?:\r|\n|$)/g, '\n');
        }
        return src;
    }
    
    
    
    
};



//        function _concatModule(moduleId, requiredMap, moduleMap, cwd, recursion) {
//			
//			if (requiredMap[moduleId]) {
//				return "";
//			}
//			requiredMap[moduleId] = 1;
//			var moduleObj = moduleMap[moduleId];
//			if (!moduleObj) {
//				throw 'moduleId(' + moduleId + ') concat error: module is not defined!';
//			}
//			
////			if (moduleObj.result) {
////				return moduleObj.result;
////			}
//			
//			var fileFullPath = moduleObj.filepath;
//			var fileFolderPath = path.dirname(fileFullPath);
//            var src = moduleObj.src;
//			var ret = [src];
//            
//            removeComment(src).replace(/\brequire\(['"](.*?)['"]\)/g, function(_, moduleId) {
//				if (/^\./.test(moduleId)) {
//					moduleId = path.join(fileFolderPath, moduleId).replace(new RegExp('^' + cwd), '').replace(/^[\\\/]/, '');
//				}
//				console.log('brequire ret.push(concatModule(m  ' + moduleId);
//				ret.push(concatModule(moduleId, requiredMap, moduleMap, cwd, true));
//            });
//			
//			var middleResult = ret.join('\n');
//			console.log('middleResult:' + middleResult[0])
//			if (recursion) {
//				console.log('caller  middleResult:' + middleResult[0])
//				return middleResult;
//			}
//			
//			var regList = [];
//			
//			removeComment(middleResult, {block: 1}).replace(/\/\/\/require\(['"](.*?)['"]\)/g, function(_, moduleReg) {
//				regList.push('(' + moduleReg + ')');
//            });
//			
//			
//			var reg = regList.length && new RegExp(regList.join('|'));
//			console.log('reg', regList, reg);
//			if (reg) {
//				if (reg.test(moduleId)) {
//					throw 'reg moduleId(' + moduleId + ') concat error://require("moduleReg") have circular reference->' + reg;
//				}
//				
//				for (var _moduleId in moduleMap) {
//					if (reg.test(_moduleId)) {
//						console.log('ret.push(concatModule(m');
//						ret.push(concatModule(moduleId, requiredMap, moduleMap, cwd, true));
//					}
//				}
//			}
////            return moduleMap[moduleId].result = ret.join('\n');
//            return ret.join('\n');
//        }
